﻿using System;
using Cs2dMapLib.IO;
using System.Collections.Generic;
using System.IO;

namespace Cs2dMapLib
{   
    /// <summary>
    /// Class for storing map data
    /// </summary>
    public class Map
    {
        /// <summary>
        /// It is not recommended to change
        /// </summary>
        public string HeaderStart;

        /// <summary>
        /// Background scrolling
        /// </summary>
        public bool BackScroling;

        /// <summary>
        /// //Unknown
        /// </summary>
        public int Check_ms;

        /// <summary>
        /// Author USGN
        /// </summary>
        public int AuthorId;

        /// <summary>
        /// Daylight Time
        /// </summary>
        public int DaylightTime;

        /// <summary>
        /// Author Nickname
        /// </summary>
        public string AuthorNick;

        /// <summary>
        /// It is not recommended to change
        /// </summary>
        public string Code;

        /// <summary>
        /// Tileset Name
        /// </summary>
        public string TilesName;

        /// <summary>
        /// Background file name
        /// </summary>
        public string BackName;

        /// <summary>
        /// Background move speed
        /// </summary>
        public XY BackScrollSpeed;

        /// <summary>
        /// Background color
        /// </summary>
        public TColor BackColor;

        /// <summary>
        /// It is not recommended to change
        /// </summary>
        public string HeaderEnd;

        /// <summary>
        /// Tileset info 
        /// </summary>
        public TileSet TSet;

        /// <summary>
        /// Tilemap info
        /// </summary>
        public TileMap TMap;

        /// <summary>
        /// Entity list
        /// </summary>
        public List<Entity> Entites;


        /// <summary>
        /// Open a map
        /// </summary>
        /// <param name="Path">The path to map</param>
        public Map(string Path)
        {
            MapReader reader = new MapReader(Path);

            HeaderStart = reader.ReadString();

            byte[] bytes = reader.ReadBytes(10);
            int[] ints = reader.ReadInts(10);
            string[] lines = reader.ReadStrings(10);

            BackScroling = bytes[0] == 0 ? false : true;
            bool map_blending = bytes[1] == 0 ? false : true;
            Check_ms = ints[0];
            AuthorId = ints[1] - 51;
            DaylightTime = ints[2];
            AuthorNick = lines[0];

            Code = reader.ReadString();
            TilesName = reader.ReadString();
            TSet = new TileSet((byte)(reader.ReadByte() + 1));
            TMap = new TileMap(reader.ReadInt() + 1, reader.ReadInt() + 1);
            BackName = reader.ReadString();
            BackScrollSpeed = new XY(reader.ReadInt(), reader.ReadInt());
            BackColor = new TColor(reader.ReadByte(), reader.ReadByte(), reader.ReadByte());

            HeaderEnd = reader.ReadString();

            

            for (int i = 0; i < TSet.Tile.Length; i++)
                TSet.Tile[i] = (TileMode)reader.ReadByte();

            for (int x = 0; x < TMap.SizeX; x++)
                for (int y = 0; y < TMap.SizeY; y++)
                    TMap.Tiles[x, y].Tile = reader.ReadByte();

            if (map_blending)
            {
                for (int x = 0; x < TMap.SizeX; x++)
                    for (int y = 0; y < TMap.SizeY; y++)
                    {
                        byte modifer = reader.ReadByte();

                        if (modifer >= 128)
                        {
                            TMap.Tiles[x, y].SetColor(reader.ReadByte(), reader.ReadByte(), reader.ReadByte());
                            TMap.Tiles[x, y].Overlay = reader.ReadByte();
                            modifer -= 128;
                        }
                        else if (modifer >= 64)
                        {
                            byte blend = reader.ReadByte();
                            BlendType type;

                            if (blend >= 32)
                            {
                                type = BlendType.Grass;
                                blend -= 32;
                            }
                            else if (blend >= 24)
                            {
                                type = BlendType.DirtH;
                                blend -= 24;
                            }
                            else if (blend >= 16)
                            {
                                type = BlendType.DirtS;
                                blend -= 16;
                            }
                            else if (blend >= 8)
                            {
                                type = BlendType.LinearH;
                                blend -= 8;
                            }
                            else
                            {
                                type = BlendType.LinearS;
                            }

                            TMap.Tiles[x, y].SetBlend(type, (BlendRotate)blend);
                            modifer -= 64;
                        }
                        if (modifer > 0)
                        {
                            string rotBits = Convert.ToString(modifer, 2);
                            if (rotBits.Length > 2)
                                rotBits = rotBits.Remove(0, rotBits.Length - 2);
                            switch (rotBits)
                            {
                                case "01":
                                    TMap.Tiles[x, y].TRotation = Rotation.Right;
                                    modifer -= 1;
                                    break;
                                case "10":
                                    TMap.Tiles[x, y].TRotation = Rotation.Down;
                                    modifer -= 2;
                                    break;
                                case "11":
                                    TMap.Tiles[x, y].TRotation = Rotation.Left;
                                    modifer -= 3;
                                    break;
                            }
                            TMap.Tiles[x, y].TBrightness = (Brightness)modifer;
                        }
                    }
            }

            Entites = new List<Entity>();
            int entityCount = reader.ReadInt();
            for (int i = 0; i < entityCount; i++)
            {
                string name = reader.ReadString();
                byte type = reader.ReadByte();
                XY pos = new XY(reader.ReadInt(), reader.ReadInt());
                string trigger = reader.ReadString();
                int[] eInts = new int[10];
                string[] eStrs = new string[10];
                for (int j = 0; j < 10; j++)
                {
                    eInts[j] = reader.ReadInt();
                    eStrs[j] = reader.ReadString();
                }
                Entites.Add(new Entity(name, (EntityType)type, pos, trigger, eInts, eStrs));
            }
            reader.Close();
        }

        /// <summary>
        /// Create new map 
        /// </summary>
        /// <param name="SizeX">Size map X</param>
        /// <param name="SizeY">Size map Y</param>
        public Map(int SizeX, int SizeY)
        {
            HeaderStart = "Unreal Software's Counter-Strike 2D Map File (max)";

            BackScroling = false;
            Check_ms = 0;
            AuthorId = 0;
            DaylightTime = 0;
            AuthorNick = "";
            Code = "4928x34$115001%11654488";
            TilesName = "cs2dnorm.bmp";
            BackName = "";
            BackScrollSpeed = new XY(0, 0);
            BackColor = new TColor(0, 0, 0);

            HeaderEnd = "ed.erawtfoslaernu";

            TSet = new TileSet(255);

            TMap = new TileMap(SizeX, SizeY);

            Entites = new List<Entity>();

        }


        /// <summary>
        /// Save a map
        /// </summary>
        /// <param name="Path">The path to save map</param>
        public void Save(string Path)
        {
            MapWriter writer = new MapWriter(Path);

            writer.Write(HeaderStart);

            byte[] bytes = new byte[10];
            int[] ints = new int[10];
            string[] lines = new string[10];

            for(int i = 0; i < 10; i++)
            {
                bytes[i] = new byte();
                ints[i] = new int();
                lines[i] = "";
            }

            bytes[0] = (byte)(BackScroling ? 1 : 0);
            bytes[1] = (byte)(TMap.BlendCheck() ? 1 : 0);
            ints[0] = Check_ms;
            ints[1] = AuthorId + 51;
            ints[2] = DaylightTime;
            lines[0] = AuthorNick;           

            writer.Write(bytes);
            writer.Write(ints);
            writer.Write(lines);

            writer.Write(Code);
            writer.Write(TilesName);
            writer.Write((byte)(TSet.Tile.Length - 1));
            writer.Write(TMap.SizeX-1);
            writer.Write(TMap.SizeY-1);
            writer.Write(BackName);
            writer.Write(BackScrollSpeed.GetArray());
            writer.Write(BackColor.GetColorBytes());
            writer.Write(HeaderEnd);

            for (int i = 0; i < TSet.Tile.Length; i++)
                writer.Write((byte)TSet.Tile[i]);

            for (int x = 0; x < TMap.SizeX; x++)
                for (int y = 0; y < TMap.SizeY; y++)
                    writer.Write(TMap.Tiles[x, y].Tile);

            if (TMap.BlendCheck())
            {
                for (int x = 0; x < TMap.SizeX; x++)
                    for (int y = 0; y < TMap.SizeY; y++)
                    {
                        writer.Write(TMap.Tiles[x, y].GetModByte());
                        if (TMap.Tiles[x, y].ModType == ModifyType.Blend)
                            writer.Write(TMap.Tiles[x,y].GetBlendByte());
                        else if (TMap.Tiles[x, y].ModType == ModifyType.Color)
                            writer.Write(TMap.Tiles[x, y].GetColorByteMassive());
                    }
            }

            writer.Write(Entites.Count);
            for (int i = 0; i < Entites.Count; i++)
            {
                writer.Write(Entites[i].Name);
                writer.Write((byte)Entites[i].Type);
                writer.Write(Entites[i].Position.GetArray());
                writer.Write(Entites[i].Trigger);
              
                for (int j = 0; j < 10; j++)
                {
                    writer.Write(Entites[i].Ints[j]);
                    writer.Write(Entites[i].Strs[j]);
                }
            }
            writer.Close();
        }
    }

    /// <summary>
    /// Class for storing tile map data
    /// </summary>
    public class TileMap
    {
        /// <summary>
        /// Size map X
        /// </summary>
        public int SizeX { get; private set; }

        /// <summary>
        /// Size map Y
        /// </summary>
        public int SizeY { get; private set; }

        /// <summary>
        /// Tile massive. Initialized when calling the constructor. ...= new TileInfo[SizeX, SizeY]
        /// </summary>
        public TileInfo[,] Tiles;


        /// <summary>
        /// Create a tiled map of size
        /// </summary>
        /// <param name="SizeX">Size map X</param>
        /// <param name="SizeY">Size map Y</param>
        public TileMap(int SizeX, int SizeY)
        {
            this.SizeX = SizeX;
            this.SizeY = SizeY;
            Tiles = new TileInfo[SizeX, SizeY];
            for (int x = 0; x < SizeX; x++)
                for (int y = 0; y < SizeY; y++)
                {
                    Tiles[x, y] = new TileInfo();
                }
        }


        /// <summary>
        /// Check if this map uses textures blend
        /// </summary>
        /// <returns></returns>
        public bool BlendCheck()
        {
            foreach(TileInfo tile in Tiles)
            {
                if (tile.BlendCheck())
                    return true;
            }
            return false;
        }
    }

    /// <summary>
    /// Class for storing tile map data
    /// </summary>
    public class TileInfo
    {
        /// <summary>
        /// Tile id
        /// </summary>
        public byte Tile;

        /// <summary>
        /// Tile rotation
        /// </summary>
        public Rotation TRotation;

        /// <summary>
        /// Tile brightness
        /// </summary>
        public Brightness TBrightness;

        /// <summary>
        /// Tile modification type
        /// </summary>
        public ModifyType ModType { get; private set; }

        /// <summary>
        /// Tile blend type
        /// </summary>
        public BlendType BType { get; private set; }

        /// <summary>
        /// Tile blend rotate
        /// </summary>
        public BlendRotate BRotate { get; private set; }

        /// <summary>
        /// Tile rgb color
        /// </summary>
        public TColor TileColor { get; private set; }

        /// <summary>
        /// Unknown
        /// </summary>
        public byte Overlay;


        /// <summary>
        /// Create tile info
        /// </summary>
        public TileInfo()
        {
            Tile = 0;
            TRotation = Rotation.Up;
            TBrightness = Brightness.b_100;
            ModType = ModifyType.None;
            BType = BlendType.LinearS;
            BRotate = BlendRotate.r_0;
            TileColor = new TColor(0, 0, 0);
            Overlay = 0;
        }


        /// <summary>
        /// Set blend tile and reset color tile
        /// </summary>
        /// <param name="Type">Tile blend type</param>
        /// <param name="Rotate">Tile blend rotate</param>
        public void SetBlend(BlendType Type, BlendRotate Rotate)
        {
            ModType = ModifyType.Blend;
            BType = Type;
            BRotate = Rotate;
        }

        /// <summary>
        /// Set color tile and reset blend tile
        /// </summary>
        /// <param name="R">Red</param>
        /// <param name="G">Green</param>
        /// <param name="B">Blue</param>
        public void SetColor(byte R, byte G, byte B)
        {
            ModType = ModifyType.Color;
            TileColor = new TColor(R, G, B);
        }

        /// <summary>
        /// Set color tile and reset blend tile
        /// </summary>
        public void ClearModify()
        {
            ModType = ModifyType.None;
            BType = BlendType.LinearS;
            BRotate = BlendRotate.r_0;
            TileColor = new TColor(0, 0, 0);
        }

        /// <summary>
        /// Clear blend, color, rotation and brightness tile
        /// </summary>
        public void ClearAllModify()
        {
            TRotation = Rotation.Up;
            TBrightness = Brightness.b_100;
            ModType = ModifyType.None;
            BType = BlendType.LinearS;
            BRotate = BlendRotate.r_0;
            TileColor = new TColor(0, 0, 0);
            Overlay = 0;
        }

        /// <summary>
        /// Check for the presence of tile modifications
        /// </summary>
        /// <returns>true - present modifications, false - no modification</returns>
        public bool BlendCheck()
        {
            if (TRotation == Rotation.Up && TBrightness == Brightness.b_100 && ModType == ModifyType.None)
                return false;
            else
                return true;
        }

        /// <summary>
        /// Return the type of tiles modification in the form of Byte
        /// </summary>
        public byte GetModByte()
        {
            return (byte)((int)ModType + (int)TBrightness + (int)TRotation);
        }

        /// <summary>
        /// Return data on the tiles blend in the form of Byte
        /// </summary>
        public byte GetBlendByte()
        {
            return (byte)((int)BType + (int)BRotate);
        }

        /// <summary>
        /// Return an array of bytes to the cell color
        /// </summary>
        /// <returns>{TileColor.R, TileColor.G, TileColor.B, Overlay}</returns>
        public byte[] GetColorByteMassive()
        {
            return new byte[4] { TileColor.R, TileColor.G, TileColor.B, Overlay };
        }
    }

    /// <summary>
    /// Class for storing the data entity
    /// </summary>
    public class Entity
    {
        /// <summary>
        /// Entity name
        /// </summary>
        public string Name;

        /// <summary>
        /// Entity type
        /// </summary>
        public EntityType Type;

        /// <summary>
        /// Entity position
        /// </summary>
        public XY Position;

        /// <summary>
        /// Entity trigger
        /// </summary>
        public string Trigger;

        /// <summary>
        /// If an entity has some parameters such as integer, they are stored in the array. For example, rotation, radiobutton index or id
        /// </summary>
        public int[] Ints;

        /// <summary>
        /// If the object has any parameters such as integer, they are stored in the array. For example, the path or line with a message
        /// </summary>
        public string[] Strs;


        /// <summary>
        /// Create new entity
        /// </summary>
        /// <param name="Name">Entity name</param>
        /// <param name="Type">Entity type</param>
        /// <param name="Position">Entity position</param>
        /// <param name="Trigger">Entity trigger</param>
        /// <param name="Ints">If an entity has some parameters such as integer, they are stored in the array. For example, rotation, radiobutton index or id</param>
        /// <param name="Strs">If the object has any parameters such as integer, they are stored in the array. For example, the path or line with a message</param>
        public Entity(string Name, EntityType Type, XY Position, string Trigger, int[] Ints, string[] Strs)
        {
            this.Name = Name;
            this.Type = Type;
            this.Position = Position;
            this.Trigger = Trigger;
            this.Ints = Ints;
            this.Strs = Strs;
        }

        /// <summary>
        /// Create new entity
        /// </summary>
        /// <param name="Name">Entity name</param>
        /// <param name="Type">Entity type</param>
        /// <param name="Position">Entity position</param>
        /// <param name="Trigger">Entity trigger</param>
        public Entity(string Name, EntityType Type, XY Position, string Trigger)
        {
            this.Name = Name;
            this.Type = Type;
            this.Position = Position;
            this.Trigger = Trigger;
            CreateMassives();
        }

        /// <summary>
        /// Create new entity
        /// </summary>
        /// <param name="Name">Entity name</param>
        /// <param name="Type">Entity type</param>
        /// <param name="Position">Entity position</param>
        public Entity(string Name, EntityType Type, XY Position)
        {
            this.Name = Name;
            this.Type = Type;
            this.Position = Position;
            Trigger = "";
            CreateMassives();
        }

        /// <summary>
        /// Create new entity
        /// </summary>
        /// <param name="Type">Entity type</param>
        /// <param name="Position">Entity position</param>
        /// <param name="Trigger">Entity trigger</param>
        public Entity(EntityType Type, XY Position, string Trigger)
        {
            Name = "";
            this.Type = Type;
            this.Position = Position;
            this.Trigger = Trigger;
            CreateMassives();
        }

        /// <summary>
        /// Create new entity
        /// </summary>
        /// <param name="Type">Entity type</param>
        /// <param name="Position">Entity position</param>
        public Entity(EntityType Type, XY Position)
        {
            Name = "";
            this.Type = Type;
            this.Position = Position;
            Trigger = "";
            CreateMassives();
        }


        /// <summary>
        /// Used in the constructors. Creates arrays with entity parameters
        /// </summary>
        private void CreateMassives()
        {
            Ints = new int[10];
            Strs = new string[10];
            for (int i = 0; i < 10; i++)
            {
                Ints[i] = 0;
                Strs[i] = "";
            }
        }
    }

    /// <summary>
    /// Class storage rgb color
    /// </summary>
    public class TColor
    {
        /// <summary>
        /// Red
        /// </summary>
        public byte R;

        /// <summary>
        /// Green
        /// </summary>
        public byte G;

        /// <summary>
        /// Blue
        /// </summary>
        public byte B;


        /// <summary>
        /// Create rgb color
        /// </summary>
        /// <param name="R">Red</param>
        /// <param name="G">Green</param>
        /// <param name="B">Blue</param>
        public TColor(byte R, byte G, byte B)
        {
            this.R = R;
            this.G = G;
            this.B = B;
        }


        /// <summary>
        /// Return an array of bytes with color
        /// </summary>
        /// <returns>RGB color byte massive</returns>
        public byte[] GetColorBytes()
        {
            return new byte[3] { R, G, B };
        }
    }

    /// <summary>
    /// Class storage 2d coordinates
    /// </summary>
    public class XY
    {
        /// <summary>
        /// X position
        /// </summary>
        public int X;

        /// <summary>
        /// Y position
        /// </summary>
        public int Y;


        /// <summary>
        /// Create 2d coordinates
        /// </summary>
        /// <param name="X">X position</param>
        /// <param name="Y">Y position</param>
        public XY(int X, int Y)
        {
            this.X = X;
            this.Y = Y;
        }


        /// <summary>
        /// Return an array of integer with 2d coordinates
        /// </summary>
        public int[] GetArray()
        {
            return new int[2] { X, Y };
        }
    }

    /// <summary>
    /// Class storage tileset. ...gfx/tiles/[name].inf
    /// </summary>
    public class TileSet
    {
        /// <summary>
        /// An array of parameters tileset tiles
        /// </summary>
        public TileMode[] Tile;


        /// <summary>
        /// Create tileset of a predetermined number of rings
        /// </summary>
        /// <param name="Count">Number of tiles</param>
        public TileSet(byte Count)
        {
            Tile = new TileMode[Count];
            for (int i = 0; i < Count; i++)
                Tile[i] = TileMode.FloorSoundless;
        }

        /// <summary>
        /// Open the tileset file on the specified path
        /// </summary>
        /// <param name="TilesetInfoFile">Path to open tileset file</param>
        public TileSet(string TilesetInfoFile)
        {
            MapReader reader = new MapReader(TilesetInfoFile);
            List<TileMode> tm = new List<TileMode>();
            try
            {
                if (reader.ReadString() == "Unreal Software's Counter-Strike 2D Tileset Informations V 2.0")
                    while (true)
                        tm.Add((TileMode)reader.ReadByte());
                else
                {
                    reader.Close();
                    throw new FileLoadException();
                }                 
            }
            catch (EndOfStreamException)
            {
                Tile = tm.ToArray();
                reader.Close();
            }           
        }


        /// <summary>
        /// Save tileset along a predetermined path
        /// </summary>
        /// <param name="Path">Path to save tileset file</param>
        public void Save(string Path)
        {
            MapWriter writer = new MapWriter(Path);
            writer.Write("Unreal Software's Counter-Strike 2D Tileset Informations V 2.0");
            foreach(TileMode tm in Tile)
            {
                writer.Write((byte)tm);
            }
            writer.Close();
        }
    }
}